namespace Heap
{
    using System;
    using System.Collections;
    using System.Reflection;
    
    class IdentityHashCode : IHashCodeProvider 
    {
        public int GetHashCode(object obj) 
        {
            return obj.GetHashCode();
        }
    }

    public class IdentityComparer : IComparer 
    {
        public int Compare(object o1, object o2) 
        { 
            return Object.ReferenceEquals(o1, o2) ? 0 : 1;
        }
    }

    public class HeapInfo
    {
        public HeapInfo()
        {
            rootObjects = new ArrayList();
            rootClasses = new ArrayList();
            objects = new Hashtable(new IdentityHashCode(), new IdentityComparer());
            structStack = new Stack();
        }
        /// <summary> Add object to the list of root objects
        /// </summary>
        /// <param name="obj">root object
        /// 
        /// </param>
        public virtual void  AddRoot(object obj)
        {
            rootObjects.Add(obj);
        }
        
        /// <summary> Add all static components of the class to the list of root objects
        /// </summary>
        /// <param name="cls">class which static components will be added to the list of root objects
        /// 
        /// </param>
        public virtual void  AddRootClass(Type cls)
        {
            rootClasses.Add(cls);
        }
        
        /// <summary> Build object closure - collect heap info
        /// </summary>
        public virtual void  Refresh()
        {
            objects.Clear();
            greyList = null;
            blackList = null;
            foreach (object obj in rootObjects)
            {
                add(obj, this);
            }
            foreach (Type cls in rootClasses)
            {
                addStaticFields(cls);
            }
            mark();
        }
        
        /// <summary> Get list of objects referencing this object
        /// </summary>
        /// <param name="obj">referened object
        /// </param>
        /// <returns> array of objects referencing this object
        /// or <code>null</code> if this object is not accessible from specified roots
        /// 
        /// </returns>
        public virtual object[] ReferencedBy(object obj)
        {
            ObjectHeader hdr = (ObjectHeader) objects[obj];
            if (hdr != null)
            {
                return hdr.referencedBy.ToArray();
            }
            return new object[0];
        }
        
        /// <summary> Get list of classes which instances reference this object
        /// </summary>
        /// <param name="obj">referened object
        /// </param>
        /// <returns> array of classes which instances reference this object
        /// in instance count decreasing order
        /// </returns>
        public virtual Instance[] ReferencedByClass(object obj)
        {
            ObjectHeader hdr = (ObjectHeader) objects[obj];
            Hashtable hash = new Hashtable();
            if (hdr != null)
            {
                foreach (object o in hdr.referencedBy) 
                {
                    Type t = o.GetType();
                    Instance i = (Instance)hash[t];
                    if (i == null) 
                    {
                        i = new Instance(t);
                        hash[t] = i;
                    }
                    i.Count += 1;
                }
            }
            Instance[] arr = new Instance[hash.Count];
            hash.Values.CopyTo(arr, 0);
            Array.Sort(arr);
            return arr;
        }
        
        /// <summary> Find links between objects matching search criteria
        /// </summary>
        /// <param name="src">condition for referencing object
        /// </param>
        /// <param name="dst">condition for referened object
        /// </param>
        /// <returns> array of Link
        /// 
        /// </returns>
        public virtual Link[] Links(Selector src, Selector dst)
        {
            ArrayList result = new ArrayList();
            for (ObjectHeader hdr = blackList; hdr != null; hdr = hdr.next)
            {
                object target = hdr.target;
                if (dst.Select(target))
                {
                    foreach (object o in hdr.referencedBy) 
                    {
                        result.Add(new Link(o, target));
                    }
                }
            }
            return (Link[])result.ToArray(typeof(Link));
        }
        
        /// <summary> Find relations between classes matching search criteria
        /// </summary>
        /// <param name="src">condition for referencing object
        /// </param>
        /// <param name="dst">condition for referened object
        /// </param>
        /// <returns> array of Relation in links count decreasing order
        /// 
        /// </returns>
        public virtual Relation[] Relations(Selector src, Selector dst)
        {
            Hashtable result = new Hashtable();
            for (ObjectHeader hdr = blackList; hdr != null; hdr = hdr.next)
            {
                object target = hdr.target;
                Type targetClass = target.GetType();
                if (dst.Select(target))
                {
                    foreach (object o in hdr.referencedBy) 
                    {
                        if (src.Select(o)) 
                        { 
                            Relation r1 = new Relation(o.GetType(), targetClass);
                            Relation r2 = (Relation)result[r1];
                            if (r2 == null) 
                            { 
                                r2 = r1;
                                result[r1] = r1;
                            }
                            r2.Count += 1;
                        }
                    }
                }
            }
            Relation[] arr = new Relation[result.Count];
            result.Keys.CopyTo(arr, 0);
            Array.Sort(arr);
            return arr;
        }
        
        /// <summary>
        /// Get information about class instances accessible from specified roots
        /// </summary> 
        /// <param name="selector">Selector of objects</param>
        /// <returns> array of Instance in count descending order</returns>
        public virtual Instance[] Instances(Selector selector)
        {
            Hashtable result = new Hashtable();
            for (ObjectHeader hdr = blackList; hdr != null; hdr = hdr.next)
            {
                object target = hdr.target;
                if (selector.Select(target))
                {
                    Type cls = target.GetType();
                    Instance inst = (Instance) result[cls];
                    if (inst == null)
                    {
                        inst = new Instance(cls);
                        result[cls] = inst;
                    }
                    inst.Count += 1;
                    if (cls.IsArray)
                    {
                        inst.TotalLength += ((Array) target).Length;
                    }
                }
            }
            Instance[] arr = new Instance[result.Count];
            result.Values.CopyTo(arr, 0);
            Array.Sort(arr);
            return arr;
        }
        
        private void  add(object obj, object from)
        {
            if (obj != null)
            {
                ObjectHeader hdr = (ObjectHeader) objects[obj];
                if (hdr == null)
                {
                    hdr = new ObjectHeader(obj);
                    hdr.next = greyList;
                    greyList = hdr;
                    objects[obj] = hdr;
                }
                hdr.referencedBy.Add(from);
            }
        }
        
        private void  addStaticFields(Type cls)
        {
            do 
            {
                foreach (FieldInfo f in getFields(cls)) {
                    if (f.IsStatic)
                    {
                        Type type = f.FieldType;
                        if (!type.IsPrimitive && !type.IsEnum)
                        {
                            if (type.IsValueType) 
                            { 
                                if (!structStack.Contains(type)) 
                                {
                                    structStack.Push(type);
                                    addStaticFields(type);
                                    structStack.Pop();
                                }
                            } 
                            else 
                            {
                                add(f.GetValue(null), cls);
                            }
                        }
                    }
                }
            }
            while ((cls = cls.BaseType) != null);
        }
        
        private void  mark()
        {
            while (greyList != null)
            {
                ObjectHeader hdr = greyList;
                greyList = hdr.next;
                hdr.next = blackList;
                blackList = hdr;
                addReferencedObjects(hdr.target);
            }
        }
        
        
        private void  addReferencedObjects(Object obj)
        {
            if (obj is object[])
            {
                object[] arr = (object[]) obj;
                for (int i = 0; i < arr.Length; i++)
                {
                    add(arr[i], obj);
                }
            }
            else if (!(obj is Type))
            {
                Type cls = obj.GetType();
                if (cls.IsValueType) 
                {
                    if (!structStack.Contains(cls)) 
                    {
                        structStack.Push(cls);
                        foreach (FieldInfo f in getFields(cls)) 
                        {
                            Type type = f.FieldType;
                            if (!type.IsPrimitive && !type.IsEnum)
                            {
                                addReferencedObjects(f.GetValue(obj));
                            }
                        }
                        structStack.Pop();
                    }
                } 
                else 
                {
                    do 
                    {
                        foreach (FieldInfo f in getFields(cls)) 
                        {
                            Type type = f.FieldType;
                            if (!type.IsPrimitive && !type.IsEnum)
                            {
                                if (f.IsStatic)
                                {
                                    add(f.GetValue(obj), cls);
                                }
                                else
                                {
                                    add(f.GetValue(obj), obj);
                                }
                            }
                        }
                    }
                    while ((cls = cls.BaseType) != null);
                }
            }
        }
        
        private FieldInfo[] getFields(Type cls)
        {
            return cls.GetFields(BindingFlags.Instance | BindingFlags.Static | BindingFlags.NonPublic | BindingFlags.Public | BindingFlags.DeclaredOnly);
        }
        
        internal class ObjectHeader
        {
            internal ObjectHeader next;
            internal bool marked;
            internal object target;
            internal ArrayList referencedBy;
            
            internal ObjectHeader(object target)
            {
                this.target = target;
                referencedBy = new ArrayList();
            }
        }
        
        private ArrayList rootObjects;
        private ArrayList rootClasses;
        private ObjectHeader greyList;
        private ObjectHeader blackList;
        private Hashtable objects;  
        private Stack structStack;
    }   
}